from .worker import Worker
from datetime import datetime, timedelta
from typing import Callable, Tuple
from threading import Thread
from collections import defaultdict
import time, itertools


"""
    TODO: Add logging for this class.
    TODO: Max workers are not defined.

    Convert task argument into list and we want args for that list also.
    Use args as list also. It will be one to one ratio reference.
    We can also have a task with contiuous load and the other task with peak load.
    This customizatio

    The self.results will contain the information with keys, and values
    which keys are the tasks and values are the workers that are
    assigned to that tasks.
"""
class TrafficManager:
    def __init__(self, tasks: list[Callable], tasks_args: list[Tuple], oper_types: list[str], **kwargs):
        self.tasks = tasks
        self.tasks_args = tasks_args
        self.oper_types = oper_types
        self.total_workers = kwargs.get("total_workers", 10)
        self.timeout = kwargs.get("timeout", 60)
        self.every_second = kwargs.get("every_second", 1)
        self.workers_per_iter = kwargs.get("workers_per_iter", 1)
        self.processes = []
        self.results = defaultdict(list)
        self.traffic_unique_id = itertools.count()

    """
        Remember that when you run the traffic manager again, it will
        clean up the  collected workers.
    """
    def _before_run(self) -> None:
        print("Runing total workers: {} for each task in {}".format(
            self.total_workers, [task.__name__ for task in self.tasks]))
        print("The each task will run total of {} seconds duration".format(self.timeout))
        self.processes = []
        self.results = defaultdict(list)

    def _generate_traffic_unique_id(self):
        return str(next(self.traffic_unique_id))

    def _generate_thread_id(self, traffic_name, traffic_id, thread_id):
        return "{}-{}-{}".format(traffic_name, traffic_id, thread_id)

    def _start_process(self, childs) -> None:
        for child in childs:
            child.start()

    #This wait function is general wait function.
    def _wait(self, childs) -> None:
        for child in childs:
            child.join()

    def get_run_result(self) -> None:
        return self.results

    """
        This is the main function that Traffic Manager. Given a list of tasks
        list of args, and operation types, there are many ways the traffic manager
        can assign to each tasks with what operations. For example, you can have
        one single task performing incremental traffic, while the other task performing
        continuous loads.
    """
    def run(self) -> list[Worker]:
        self._before_run()
        for task, task_args, oper_type in zip(self.tasks, self.tasks_args, self.oper_types):
            traffic_id = self._generate_traffic_unique_id()
            print("Generated ID: {} for traffic typ: {}".format(traffic_id, oper_type))
            if oper_type == "incremental_traffic":
                p = Thread(
                    target=self.incremental_traffic,
                    args=(oper_type, traffic_id, task, task_args))
            elif oper_type == "peak_load":
                p = Thread(
                    target=self.peak_load,
                    args=(oper_type, traffic_id, task, task_args))
            elif oper_type == "continuous_load":
                p = Thread(
                    target=self.continuous_load,
                    args=(oper_type, traffic_id, task, task_args))
            else:
                raise Exception(
                    "The operation : {} couldn't be found under TrafficManager".format(
                        oper_type)
                    )
            self.processes.append(p)

        print("Starting the process")
        self._start_process(self.processes)

        print("Waiting for all processes to finish by traffic manager..")
        self._wait(self.processes)

        print("The task has finished by Traffic Manager.")
        return self.get_run_result()

    """
        Perform incremental traffic and return those results in
        a list of workers. The way incremental traffic work is by 
        sending 1 client to N clients in a group one by one. The first
        test will be 1 Client, second iteration test will be 2 clients, 
        and so forth until N clients for Nth iteration.
    """
    def incremental_traffic(self, traffic_name, traffic_id, task, task_args) -> None:
        result = []
        id_tracker = 0
        current_workers = 1
        end_duration_time = datetime.now() + timedelta(seconds=self.timeout)

        while datetime.now() < end_duration_time:
            workers = []
            print("Setting total of {} workers".format(current_workers))
            for _ in range(current_workers):
                workers.append(
                    Worker(
                        id=self._generate_thread_id(
                            traffic_name, traffic_id, id_tracker),
                        args=task_args,
                        task=task
                    )
                )
                id_tracker += 1

            for worker in workers:
                worker.start()
            
            self._wait(workers)
            result += workers

            if current_workers < self.total_workers:
                current_workers += 1
        
        self.results[task.__name__] += result

    """
        Run all N clients at the same time for the certain duration.
        Each of the worker will do the task again and again until the
        timout of the test duration.
    """
    def peak_load(self, traffic_name, traffic_id, task, task_args) -> None:
        workers = []
        id_tracker = 0

        for _ in range(self.total_workers):
            workers.append(
                Worker(
                    id=self._generate_thread_id(
                        traffic_name, traffic_id, id_tracker
                    ),
                    task=task,
                    timeout=self.timeout,
                    repeat=True,
                    args=task_args
                )
            )
            id_tracker += 1

        for worker in workers:
            worker.start()
        
        print("Scheduled workers, just waiting for workers to finish")
        self._wait(workers)

        self.results[task.__name__] += workers

    """
        Continuous load will send X requests (or generate clients/workers)
        per n amount of seconds. For more furhter details, please check
        the documentation.
    """
    def continuous_load(
        self,
        traffic_name: str,
        traffic_id,
        task,
        task_args
    ) -> None:
        end_duration_time = datetime.now() + timedelta(seconds=self.timeout)
        workers = []
        id_tracker = 0

        while datetime.now() < end_duration_time:
            this_iter_workers = []
            for _ in range(self.workers_per_iter):
                this_iter_workers.append(
                    Worker(
                        id=self._generate_thread_id(
                            traffic_name, traffic_id, id_tracker
                        ),
                        task=task,
                        args=task_args
                    )
                )
                id_tracker += 1

            for worker in this_iter_workers:
                worker.start()
            workers += this_iter_workers
            time.sleep(self.every_second)

        self._wait(workers)

        self.results[task.__name__] += workers

    


from mago.utils.collector import Collector
from unittest.mock import MagicMock
from mago.utils.worker import Worker
from datetime import datetime, timedelta
import time

def task_test():
    pass

test_data = {task_test.__name__: []}
timer = datetime.now()

for id in range(10):
    worker = Worker(id=id, task=task_test, args=())
    worker.get_result = MagicMock(
        return_value={
            "id": id,
            "log": "Worker Log {}".format(id),
            "created": datetime.now().timestamp(),
            "result": [{
                "start_time": timer.timestamp(),
                "end_time": (timer + timedelta(seconds=5)).timestamp(),
                "duration": 5,
                "value": 5,
                "status": "SUCCESS"
            }]
        }
    )
    worker.get_total_tasks_executed = MagicMock(
        return_value=1
    )
    timer += timedelta(seconds=1)
    test_data[task_test.__name__].append(worker)

def test_workers_result():
    results = Collector.collect_workers_result(test_data)
    assert task_test.__name__ in results
    assert len(results[task_test.__name__]) == 10
    for id in range(len(results[task_test.__name__])):
        each = results[task_test.__name__][id]
        assert each['start_time'] < each['end_time']
        assert each['duration'] == 5
        assert each['value'] == 5
        assert each['status'] == "SUCCESS"
        assert each['id'] == id

def test_collect_total_clients_per_timeline():
    results = Collector.collect_total_clients_per_timeline(test_data)
    assert len(results) == 1
    assert len(results[0]) > 1
    for each in results:
        assert each['label'] == task_test.__name__
        assert each['tension'] == 0.5
        assert len(each['data']) > 0
        for each_data in each['data']:
            assert 'x' in each_data
            assert 'y' in each_data

def test_total_task_finished_per_worker():
    x_labels, y_labels = Collector.collect_total_task_finished_per_worker(test_data)
    assert len(x_labels)
    assert len(y_labels) == 1
    for each in y_labels:
        assert each['label'] == task_test.__name__
        assert each['tension'] == 0.5
        assert len(each['data']) > 0
        for each_data in each['data']:
            assert each_data == 1
    
def test_collect_percentile():
    results = Collector.collect_percentile(test_data)
    assert len(results) == 1
    for each in results:
        assert len(each) == 5
        assert each[0] == task_test.__name__
        assert each[1] > 0
        assert each[2] > 0
        assert each[3] > 0
        assert each[4] > 0

def test_give_average_durations():
    min_time, y_labels = Collector.give_average_durations(test_data)
    assert type(min_time) == str
    assert len(y_labels) == 1
    for each in y_labels:
        assert each['label'] == task_test.__name__
        assert each['tension'] == 0.5
        assert len(each['data']) > 0
        for each_data in each['data']:
            assert 'x' in each_data
            assert 'y' in each_data


    


from .utils.report_generator import ReportGenerator
from .utils.traffic_manager import TrafficManager
from .utils.collector import Collector
from typing import Callable, Tuple
import logging, os


logging.basicConfig(level=logging.DEBUG)
"""
    This is the main Seshat which users will interact for their
    own custom testing method.

    TODO: Create a test for Seshat itself
"""
class Mago:
    def __init__(
        self,
        tasks: list[Callable],
        tasks_args: list[Tuple],
        oper_types: list[str],
        total_workers: int,
        output_path: str,
        timeout: int = 60
    ):
        self.output_path = output_path
        if not os.path.exists(self.output_path):
            os.mkdir(self.output_path)

        self.traffic_manager = TrafficManager(
            tasks=tasks,
            tasks_args=tasks_args,
            oper_types=oper_types,
            total_workers=total_workers,
            timeout=timeout
        )
        self.report_generator = ReportGenerator(output=output_path)
        self.logger = logging.getLogger(__name__)

    """
        This function will store the result.
    """
    def store_result(self, data: dict):
        self.report_generator.write_home_content(data)
        self.report_generator.write_workers_content(data)
        self.report_generator.write_log_content(data)

    def collect_data(self, results: dict[str, list]) -> dict:
        data = {}
        data['num_of_workers_datasets'] =\
            Collector.collect_total_clients_per_timeline(results)

        data['avg_resp_min_time'], data['avg_response_datasets'] =\
            Collector.give_average_durations(results)

        data['num_of_tasks_x_labels'], data['num_of_tasks_datasets'] =\
            Collector.collect_total_task_finished_per_worker(results)

        data['task_distributed_labels'], data['task_distributed_percentage'] =\
            Collector.get_task_distributed(results)

        data['tasks_passed_labels'], data['task_passed_data'] =\
            Collector.get_tasks_passed(results)

        data['polar_area_labels'], data['polar_area_average_data'] =\
            Collector.get_average_time_tasks(results)

        data['percentile_data_table'] = Collector.collect_percentile(results)
        data['passFailLabels'], data['passFailBarData'] =\
            Collector.get_pass_fail_percentage(results)

        data['logDataInfo'] = Collector.collect_log_data(results)

        return data

    def run(self):
        self.traffic_manager.run()
        results = self.traffic_manager.get_run_result()
        finalized_data = self.collect_data(results)
        self.store_result(finalized_data)


from .worker import Worker
from collections import defaultdict
import numpy as np


class Collector:
    @staticmethod
    def generate_datasets(y_labels: dict[str, list]):
        datasets = []
        for task, data in y_labels.items():
            temp = {}
            temp['label'] = task
            temp['tension'] = 0.5
            temp['data'] = data
            datasets.append(temp)
        return datasets

    @staticmethod
    def collect_workers_result(data: dict[str, list[Worker]]) -> dict[str, list[dict]]:
        results = defaultdict(list)
        for task, workers in data.items():
            for worker in workers:
                result = worker.get_result()
                for each_exec in result['result']:
                    temp = {}
                    temp.update({
                        "id": result['id']
                    })
                    temp.update(each_exec)
                    results[task].append(temp)
        return results

    @staticmethod
    def collect_total_clients_per_timeline(data: dict[str, list[Worker]], time_increment=1):
        new_data = Collector.collect_workers_result(data)

        count_start_and_end_time = {}

        for task, each_data in new_data.items():
            count_start_and_end_time[task] = defaultdict(int)
            for exec_data in each_data:
                rounded_start_time = round(exec_data['start_time'], 0)
                rounded_end_time = round(exec_data['end_time'], 0)
                count_start_and_end_time[task][rounded_start_time] += 1
                count_start_and_end_time[task][rounded_end_time] -= 1

        y_labels = {}
        for task in count_start_and_end_time:
            y_labels[task] = []
            sorted_time = [*count_start_and_end_time[task]]
            sorted_time.sort()
            total_active_workers = 0

            for time in sorted_time:
                total_active_workers += count_start_and_end_time[task][time]
                y_labels[task].append({
                    'x': time,
                    'y': total_active_workers
                })
            
        return Collector.generate_datasets(y_labels)

    @staticmethod
    def collect_total_task_finished_per_worker(data: dict[str, list[Worker]]):
        index = 0
        x_labels = []
        y_labels = {}
        indices = defaultdict()

        """
        generate key, value which the key is thread_unique_ID
        #and the value is the index that will point within the list.
        for example, two unique thread ID will have
        {
            "threadID#1": 0,
            "threadID#2": 1
        }
        Here index will also count how many total workers are there.
        """
        for _, each_data in data.items():
            for worker in each_data:
                indices[worker._id] = index
                x_labels.append(worker._id)
                index += 1

        for task, each_data in data.items():
            y_labels[task] = [0 for _ in range(index)]
            for worker in each_data:
                y_labels[task][indices[worker._id]] = worker.get_total_tasks_executed()
        
        return x_labels, Collector.generate_datasets(y_labels)

    @staticmethod
    def collect_percentile(data: dict[str, list[Worker]]):
        percentile_data_table = []
        for task, workers in data.items():
            duration_data = []
            for worker in workers:
                durations = [e['duration'] for e in worker.get_result()['result']]
                duration_data += durations

            np_array = np.array(duration_data)
            percentile_data_table.append([
                task,
                np.percentile(np_array, 99),
                np.percentile(np_array, 99.9),
                np.percentile(np_array, 99.99),
                np.percentile(np_array, 99.999)
            ])
        return percentile_data_table

    # TODO: This is pretty complicated, we need to break down into simpler terms.
    @staticmethod
    def give_average_durations(data: dict[str, list[Worker]]):
        formatted_data = defaultdict(list)
        min_time = None
        for task, workers in data.items():
            for worker in workers:
                for exec_result in worker.get_result()['result']:
                    formatted_data[task].append([
                        exec_result['end_time'], exec_result['duration']
                    ])
                        
                    if min_time == None:
                        min_time = exec_result['start_time']
                    else:
                        min_time = min(min_time, exec_result['start_time'])

        collected_data = {}

        for task, times in formatted_data.items():
            collected_data[task] = {}
            for end_time, duration in times:
                if not end_time in collected_data[task]:
                    collected_data[task][end_time] = [duration, 1]
                else:
                    collected_data[task][end_time][0] += duration
                    collected_data[task][end_time][1] += 1

        y_labels = {}
        for task in collected_data:
            y_labels[task] = []
            for time in collected_data[task]:
                y_labels[task].append({
                    'x': time,
                    'y': collected_data[task][time][0] / collected_data[task][time][1] 
                })
            y_labels[task].sort(key=lambda d: d['x'])
        
        return str(min_time), Collector.generate_datasets(y_labels)

    @staticmethod
    def get_task_distributed(data: dict[str, list[Worker]]):
        results = defaultdict(int)
        for task, workers in data.items():
            for worker in workers:
                results[task] += worker.get_total_tasks_executed()
        
        task_distributed_labels = list(results.keys())
        task_distributed_percentage = [
            results[task] for task in task_distributed_labels]

        return task_distributed_labels, task_distributed_percentage

    @staticmethod
    def get_tasks_passed(data: dict[str, list[Worker]]):
        total_passed = defaultdict(int)
        total_executed = defaultdict(int)
        for task, workers in data.items():
            for worker in workers:
                total_passed[task] += worker.get_total_passed()
                total_executed[task] += worker.get_total_tasks_executed()

        tasks_passed_labels = list(data.keys())
        task_passed_data = []
        for task in tasks_passed_labels:
            task_passed_data.append(
                round(total_passed[task]/total_executed[task] * 100, 2))

        return tasks_passed_labels, task_passed_data

    @staticmethod
    def get_average_time_tasks(data: dict[str, list[Worker]]):
        results = defaultdict(list)
        for task, workers in data.items():
            for worker in workers:
                for each_task in worker.get_result()['result']:
                    results[task].append(each_task['duration'])

        polar_area_labels = list(results.keys())
        polar_area_average_data = [
            sum(results[task])/len(results[task]) for task in results]

        return polar_area_labels, polar_area_average_data

    @staticmethod
    def get_pass_fail_percentage(data: dict[str, list[Worker]]):
        datasets = []
        labels = []
        passed_data = []
        failed_data = []
        timedout_data = []
        for task, workers in data.items():
            total_passed = total_failed = total_timedout = 0
            for worker in workers:
                total_passed += worker.get_total_passed()
                total_failed += worker.get_total_failed()
                total_timedout += worker.get_total_timedout()

            passed_data.append(total_passed)
            failed_data.append(total_failed)
            timedout_data.append(total_timedout)

            labels.append(task)

        datasets.append({
            'label': "Passed",
            'data': passed_data,
            'backgroundColor': 'green'
        })

        datasets.append({
            'label': "Failed",
            'data': failed_data,
            'backgroundColor': 'red'
        })

        datasets.append({
            'label': "TimedOut",
            'data': timedout_data,
            'backgroundColor': 'yellow'
        })

        return labels, datasets

    @staticmethod
    def collect_log_data(data: dict[str, list[Worker]]):
        datasets = []
        for task, workers in data.items():
            for worker in workers:
                datasets.append(
                    [
                        worker._id,
                        task,
                        worker.get_total_tasks_executed(),
                        worker.get_total_passed(),
                        worker.get_total_failed(),
                        worker.get_total_min_duration(),
                        worker.get_total_max_duration(),
                        worker.get_avg_duration(),
                        worker.get_total_timedout()
                    ]
                )
        return datasets



#!/usr/bin/env python3
import json
import subprocess


def get_last_version() -> str:
    """Return the version number of the last release."""
    json_string = (
        subprocess.run(
            ["gh", "release", "view", "--json", "tagName"],
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        .stdout.decode("utf8")
        .strip()
    )

    return json.loads(json_string)["tagName"]


def bump_patch_number(version_number: str) -> str:
    """Return a copy of `version_number` with the patch number incremented."""
    major, minor, patch = version_number.split(".")
    return f"{major}.{minor}.{int(patch) + 1}"


def create_new_patch_release():
    """Create a new patch release on GitHub."""
    try:
        last_version_number = get_last_version()
    except subprocess.CalledProcessError as err:
        if err.stderr.decode("utf8").startswith("HTTP 404:"):
            # The project doesn't have any releases yet.
            new_version_number = "0.0.1"
        else:
            raise
    else:
        new_version_number = bump_patch_number(last_version_number)

    subprocess.run(
        ["gh", "release", "create", "--generate-notes", new_version_number],
        check=True,
    )


if __name__ == "__main__":
    create_new_patch_release()


from mago.utils.traffic_manager import TrafficManager
import random, time


def test_traffic_manager_initalization():
    def print_hello_world():
        pass

    tasks = [print_hello_world]
    tasks_args = [()]
    oper_types = ["peak_load"]
    tm = TrafficManager(
        tasks=tasks,
        tasks_args=tasks_args,
        oper_types=oper_types
    )

    assert tm.tasks == tasks
    assert tm.tasks[0].__name__ == "print_hello_world"
    assert tm.total_workers == 10
    assert tm.timeout == 60

def test_peak_load():
    def random_sleep():
        sleep_time = random.randrange(1,3)
        time.sleep(sleep_time)

    tm = TrafficManager(
        tasks=[random_sleep],
        tasks_args=[()],
        oper_types=["peak_load"],
        timeout=10
    )

    assert tm.timeout == 10

    tm.run()
    results = tm.get_run_result()
    
    assert results['random_sleep'] != None

    workers = results['random_sleep']
    assert len(workers) == 10

    unique_id = set()

    for worker in workers:
        assert not worker._id in unique_id
        unique_id.add(worker._id)
        result = worker.get_result()
        assert 'peak_load' in result['id']
        assert result['log'] != None
        assert len(result['result']) >= 2

        for exec_result in result['result']:
            assert exec_result['end_time'] > exec_result['start_time']

def test_incremental_traffic():
    def sleep_three_seconds():
        time.sleep(3)

    tm = TrafficManager(
        tasks=[sleep_three_seconds],
        tasks_args=[()],
        oper_types=["incremental_traffic"]
    )

    tm.run()
    workers = tm.get_run_result()["sleep_three_seconds"]

    assert len(workers) == 55
    unique_id = set()

    for worker in workers:
        assert not worker._id in unique_id
        unique_id.add(worker._id)
        result = worker.get_result()
        assert 'incremental_traffic' in result['id']
        assert result['log'] != None
        assert len(result['result']) == 1
        assert result['created'] != None

        for exec_result in result['result']:
            assert exec_result['end_time'] > exec_result['start_time']

def test_continuous_load():
    def sleep_three_seconds():
        time.sleep(3)

    tm = TrafficManager(
        tasks=[sleep_three_seconds],
        tasks_args=[()],
        oper_types=["continuous_load"],
        timeout=10,
        every_second=1,
        workers_per_iter=3
    )

    tm.run()
    workers = tm.get_run_result()['sleep_three_seconds']
    unique_id = set()

    for worker in workers:
        assert not worker._id in unique_id
        unique_id.add(worker._id)
        result = worker.get_result()
        assert 'continuous_load' in result['id']
        assert result['log'] != None
        assert len(result['result']) == 1
        assert result['created'] != None

        for exec_result in result['result']:
            assert exec_result['end_time'] > exec_result['start_time']
            
def test_two_tasks():
    def sleep_three_seconds():
        time.sleep(3)

    def sleep_two_seconds():
        time.sleep(2)

    tm = TrafficManager(
        tasks=[sleep_two_seconds, sleep_three_seconds],
        tasks_args=[(), ()],
        oper_types=['peak_load', 'peak_load'],
        timeout=15
    )

    tm.run()
    result = tm.get_run_result()
    three_second_workers = result['sleep_three_seconds']
    two_second_workers = result['sleep_two_seconds']

    assert len(three_second_workers) == 10
    assert len(two_second_workers) == 10

def test_two_tasks_with_different_methods():
    def sleep_three_seconds():
        time.sleep(3)

    def sleep_two_seconds():
        time.sleep(2)

    tm = TrafficManager(
        tasks=[sleep_two_seconds, sleep_three_seconds],
        tasks_args=[(), ()],
        oper_types=['peak_load', 'incremental_traffic'],
        timeout=15
    )

    tm.run()
    result = tm.get_run_result()
    assert result['sleep_three_seconds'] != None
    assert result['sleep_two_seconds'] != None





    
    


from datetime import datetime
from threading import Thread
from typing import Callable, Tuple
from io import StringIO
import logging, time


class Worker(Thread):
    def __init__(self,
        id: str,
        task: Callable,
        args: Tuple,
        repeat: bool = False,
        timeout: int = None
    ):
        Thread.__init__(self)
        self._id = id
        self._log = ""
        self._task = task
        self._args = args
        self._timeout: int = timeout
        if self._timeout == None:
            self._repeat = False
        else:
            self._repeat: bool = repeat
        self._results = []
        self._stop_loop = False
        self._time_created = None
        self._tasks_executed = 0
        self._total_passed = 0
        self._total_timedout = 0
        self.log_stream = StringIO()
        self.log_handler = logging.StreamHandler(self.log_stream)
        self.logger = logging.getLogger("WorkerLogger:{}".format(self._id))
        self.logger.setLevel(logging.ERROR)
        self.logger.addHandler(self.log_handler)

    """
        This is the main run function that will execute the task given.
        In order to run this, please run Worker.start(). That will
        kick off this function as this function is override of Thread.
        Just note that Thread can be run only once.
    """
    def run(self) -> None:
        self._time_created = time.time()

        while not self._stop_loop:
            self.logger.info("Starting worker: {} with task: {}".format(
                self._id, self._task.__name__))
            time_started = time.time()
            status = "SUCCESS"
            result = None
            try:
                result = self._task(*self._args)
                if result == None:
                    result = "EMPTY RESPONSE"
                self._total_passed += 1
            except Exception as err:
                self.logger.error("Failed to execute due to error : {}".format(err))
                status = "FAILED"
                self._total_timedout += 1
            finally:
                time_ended = time.time()
                self.logger.info("Worker {} has succesfully finished the process.".format(
                    self._id))

                if self._timeout != None:
                    if time_ended - self._time_created > self._timeout:
                        status = "FAILED DUE TO TIMEOUT"
                        self._stop_loop = True
                
                self._results.append({
                    "start_time": time_started * 1000,
                    "end_time": time_ended * 1000,
                    "duration": (time_ended - time_started) * 1000,
                    "value": result,
                    "status": status
                })
                self._tasks_executed += 1

            if not self._repeat:
                break
        
    """
        Get all info
    """
    def get_result(self) -> dict:
        return {
            "id": self._id,
            "log": self._log,
            "created": self._time_created,
            "result": self._results
        }

    def get_total_tasks_executed(self) -> int:
        return self._tasks_executed

    def get_total_passed(self) -> int:
        return self._total_passed

    def get_total_failed(self) -> int:
        return self._tasks_executed - self._total_passed

    def get_pass_percentage(self) -> float:
        return self._total_passed/self._tasks_executed

    def get_total_timedout(self) -> int:
        return self._total_timedout

    def get_total_min_duration(self) -> float:
        return min([each_result['duration'] for each_result in self._results])

    def get_total_max_duration(self) -> float:
        return max([each_result['duration'] for each_result in self._results])

    def get_avg_duration(self) -> float:
        totalSum = sum([each_result['duration'] for each_result in self._results])
        return totalSum / self._tasks_executed

    def get_log(self) -> str:
        logs = self.log_stream.getvalue()
        if not logs:
            return "No Errors"
        return logs



from mago.utils.worker import Worker
from datetime import datetime, timedelta
import time
import pytest


@pytest.mark.parametrize("func, data, expect", [
    (sum, [1,2,3,4], 10)])
def test_worker_get_result(func, data, expect):
    test_worker = Worker(id='sum-0-0', task=func, args=(data, ))
    test_worker.start()
    test_worker.join()
    result = test_worker.get_result()
    assert result['log'] != None
    assert result['id'] == 'sum-0-0'
    assert type(result['result']) == list
    assert len(result['result']) >= 1
    assert result['result'][0]['value'] == expect
    assert result['result'][0]['status'] == "SUCCESS"
    assert result['result'][0]['start_time'] != None
    assert result['result'][0]['end_time'] != None
    assert result['result'][0]['end_time'] > result['result'][0]['start_time']

def test_worker_timeout():
    def time_sleep():
        time.sleep(5)
    
    test_worker = Worker(
        id=1, task=time_sleep, args=(), timeout=2
    )
    test_worker.start()
    test_worker.join()
    result = test_worker.get_result()
    assert result['result'][0]['status'] == "FAILED DUE TO TIMEOUT"

def test_worker_failure():
    def cause_failure():
        raise Exception("Causing test failure")

    test_worker = Worker(
        id=1, task=cause_failure, args=())

    test_worker.start()
    test_worker.join()
    result = test_worker.get_result()
    assert result['result'][0]['status'] == 'FAILED'

def test_worker_with_repeat_and_timeout():
    def sleep_one_second():
        time.sleep(1)

    test_worker = Worker(
        id=1,
        task=sleep_one_second,
        args=(),
        timeout=10,
        repeat=True)

    test_worker.start()
    test_worker.join()
    result = test_worker.get_result()
    assert len(result['result']) >= 9
    assert len(result['result']) <= 11

def test_worker_logging():
    def sleep_one_second():
        time.sleep(1)

    workers = []
    workers.append(
        Worker(
            id=1,
            task=sleep_one_second,
            args=(),
            timeout=5,
            repeat=True
        )
    )
    workers.append(
        Worker(
            id=2,
            task=sleep_one_second,
            args=(),
            timeout=5,
            repeat=True
        )
    )

    for worker in workers:
        worker.start()

    for worker in workers:
        worker.join()

    logs = [w.get_log() for w in workers]
    for log in logs:
        assert log != ""



import os

PATH_TO_TEMPLATE = os.path.join(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
    "template")

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [PATH_TO_TEMPLATE],
    }
]

from django.conf import settings
import django

settings.configure(TEMPLATES=TEMPLATES)
django.setup()

from django.template.loader import get_template

"""
    TODO: Create a test for report_generator.py
"""
class ReportGenerator:
    def __init__(self, output: str):
        self.output_folder = output
        # self.timestamp = datetime.now()
        self.template_filepath = os.path.join(
            os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
            "template")

    """
        TODO: We have to move the rest of the logo to the 
        output folder.
    """
    def write_workers_content(self, content_data: dict):
        template = get_template('workers.html')
        filepath = os.path.join(self.output_folder, "workers.html")
        
        with open(filepath, 'w') as fp:
            fp.write(template.render(content_data))

    def write_home_content(self, content_data: dict = {}):
        template = get_template("index.html")
        filepath = os.path.join(self.output_folder, "index.html")

        with open(filepath, 'w') as fp:
            fp.write(template.render(content_data))

    def write_log_content(self, content_data: dict = {}):
        template = get_template("logs.html")
        filepath = os.path.join(self.output_folder, "logs.html")
        
        with open(filepath, 'w') as fp:
            fp.write(template.render(content_data))

        